/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SystemUtil.java 446 2006-05-19 14:13:50Z ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.util;

import java.io.File;
import java.net.JarURLConnection;
import java.net.URI;
import java.net.URL;

/**
 * This class is used to locate the Petals install directory
 * 
 * @version $Rev: 446 $ $Date: 2006-05-19 16:13:50 +0200 (ven, 19 mai 2006) $
 * @since Petals 1.0
 * @author Rafael Marins
 * @author ddesjardins - eBMWebsourcing
 */
public final class SystemUtil {

    /**
     * Petals install directory
     */
    private static File petalsInstallDirectory;

    private SystemUtil() {
        super();
    }

    /**
     * Find the petals install directory
     * 
     * @return Petals install directory
     */
    public static File getPetalsInstallDirectory() {
        if (petalsInstallDirectory == null) {
            // first try to resolve install directory set in the environment
            File dir = resolveInstallDirectoryFromEnvironment();
            // otherwise, guess install directory using the framework
            if (dir == null) {
                dir = resolveInstallDirectoryFromFramework();
            }
            petalsInstallDirectory = dir;
        }
        return petalsInstallDirectory;
    }

    /**
     * Try to get the petals'home environment variable : petals.home system
     * property, or PETALS_HOME env variable
     * 
     * @return the related directory, or null
     */
    protected static File resolveInstallDirectoryFromEnvironment() {
        File directory = null;
        String petalsHome = System.getProperty("petals.home");
        if (petalsHome == null || petalsHome.trim().equals("")) {
            petalsHome = System.getenv("PETALS_HOME");
        }
        // not null or empty
        if (petalsHome != null && !petalsHome.trim().equals("")) {
            File theHome = new File(petalsHome);
            // exists and it's a directory
            if (theHome.exists() || theHome.isDirectory()) {
                directory = theHome;
            }
        }
        return directory;
    }

    /**
     * Try to find the install directory from the server.jar
     * 
     * @return the directory or null
     */
    protected static File resolveInstallDirectoryFromFramework() {
        // guess from the location of the server.jar using a marker
        URL url = SystemUtil.class.getClassLoader().getResource(
                "META-INF" + File.separator + "LICENSE.txt");
        File directory = null;
        if (url != null) {
            try {
                JarURLConnection jarConnection = (JarURLConnection) url
                        .openConnection();
                URL resourceJar = jarConnection.getJarFileURL();
                // resolve parent directory for ${basedir}/bin/server.jar
                URI baseURI = new URI(resourceJar.toString()).resolve("..");
                directory = new File(baseURI);
            } catch (Exception ignored) {
                // ignored
            }
        } else {
            // Failed to load resource \"META-INF/petals-jar
            // installPath not found
        }
        return directory;
    }

}
