/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLUtil.java 98 2006-02-24 16:18:48Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.util;

import java.io.ByteArrayInputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * Contains utilities methods for XML operations.
 * 
 * @version $Rev: 98 $ $Date: 2006-02-24 17:18:48 +0100 (ven., 24 févr. 2006) $
 * @since Petals 1.0
 * @author Adrien LOUIS
 * @author ddesjardins - eBMWebsourcing
 */
public final class XMLUtil {

    private XMLUtil() {
        // Do nothing
    }

    /**
     * Create an attribute node with the specified value
     * 
     * @param document
     *            XML document
     * @param att
     *            attribute name
     * @param value
     *            attribute value
     * @return an attribute, null if the document is null
     */
    public static Node createAttribute(Document document, String att,
        String value) {
        Node element = null;
        if (document != null) {
            element = document.createAttribute(att);
            element.setNodeValue(value);
        }
        return element;
    }

    /**
     * Create a node with the specified name, attributes and values. attVal is a
     * suite : att1,val1,att2,val2,...
     * 
     * @param document
     *            xlm document
     * @param nodeName
     *            node name
     * @param attVal
     *            attribute values
     * @return the node with its attribute, null if the array is not correct or
     *         the document is null.
     */
    public static Node createNode(Document document, String nodeName,
        String... attVal) {
        Node element = null;
        if (document != null && attVal.length % 2 == 0) {
            element = document.createElement(nodeName);
            for (int i = 0; i < attVal.length; i = i + 2) {
                element.getAttributes().setNamedItem(
                    createAttribute(document, attVal[i], attVal[i + 1]));
            }
        }
        return element;
    }

    /**
     * Return the value of the attribute in the node
     * 
     * @param n
     *            the node
     * @param attName
     *            the name of the attribute
     * @return the value of the attribute, null if not found
     */
    public static String getAttributeValue(Node n, String attName) {
        if (n != null) {
            NamedNodeMap atts = n.getAttributes();
            Node att = atts.getNamedItem(attName);
            if (att != null) {
                return att.getNodeValue();
            }
        }
        return null;
    }

    /**
     * Return the first child of a node, regardless <i>text</i> node
     * 
     * @param node
     * @return
     */
    public static Node getFirstChild(Node node) {
        Node result = node.getFirstChild();
        if (result != null) {
            while (result.getNodeType() == Node.TEXT_NODE) {
                result = result.getNextSibling();
            }
        }
        return result;
    }

    /**
     * Return the next sibling of a node, regardless <i>text</i> node
     * 
     * @param node
     * @return
     */
    public static Node getNextSibling(Node node) {
        Node result = node.getNextSibling();
        if (result != null) {
            while (result.getNodeType() == Node.TEXT_NODE) {
                result = result.getNextSibling();
            }
        }
        return result;
    }

    /**
     * return the node corresponding to the given path for the Document.
     * 
     * @param path
     * @param document
     * @return the first matching node, null otherwise, or if document/path is
     *         null
     */
    public static Node getNode(Document document, String path) {
        Node result = null;
        if (document != null && path != null) {
            try {
                result = XPathAPI.selectSingleNode(document, path);
            } catch (TransformerException e) {
                return null;
            }
        }
        return result;
    }

    /**
     * Return the <i>text</i> element of a node
     * 
     * @param node
     * @return
     */
    public static String getTextContent(Node node) {
        NodeList list = node.getChildNodes();
        for (int i = 0; i < list.getLength(); i++) {
            if (list.item(i).getNodeType() == Node.TEXT_NODE) {
                return list.item(i).getNodeValue();
            }
        }
        return null;
    }

    /**
     * Create a String result from a DOM document
     * 
     * @param document
     *            the DOM Document
     * @return a String representation of the DOM Document
     * @throws TransformerException
     */
    public static String createStringFromDOMDocument(Document document)
        throws TransformerException {
        document.normalize();
        Source source = new DOMSource(document);
        StringWriter out = new StringWriter();
        Result resultStream = new StreamResult(out);
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer;
        transformer = tFactory.newTransformer();
        transformer.transform(source, resultStream);
        String result = out.toString();

        return result;
    }

    /**
     * Create a document from a String
     * 
     * @param xml
     *            xml string
     * @return document representing the document
     */
    public static Document createDocumentFromString(String xml) {
        Document doc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                .newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            byte[] msgByte = xml.getBytes();
            ByteArrayInputStream in = new ByteArrayInputStream(msgByte);
            InputSource inputSource = new InputSource(in);
            doc = builder.parse(inputSource);
        } catch (Exception e) {
            // Do nothing
        }
        return doc;
    }

    /**
     * Return an array of String representing each Text element of the nodes
     * which are in the list.
     * 
     * @param list
     * @return
     */
    public static List<String> getTextContents(NodeList list) {
        List<String> result = null;
        if (list != null) {
            result = new ArrayList<String>();
            for (int i = 0; i < list.getLength(); i++) {
                Node pathElement = list.item(i);
                result.add(getTextContent(pathElement));
            }
        }
        return result;
    }

}
